@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2003, 2004, 2005, 2007, 2008 Free Software Foundation, Inc.
+   Copyright (C) 2003-2020 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -12,32 +12,33 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General
-   Public License along with this library; if not, write to the
-   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-   Boston, MA 02110-1301 USA */
+   Public License along with this library.  If not, see 
+   <http://www.gnu.org/licenses/>. */
 
 #ifndef _MAILUTILS_GSASL_H
 #define _MAILUTILS_GSASL_H
+#include <mailutils/cli.h>
 
 struct mu_gsasl_module_data
 {
+  int enable;
   char *service;
   char *realm;
   char *hostname;
   char *anon_user;
   char *cram_md5_pwd;
 };
+extern struct mu_gsasl_module_data mu_gsasl_module_data;
 
-int mu_gsasl_module_init (void *data);
+int mu_gsasl_enabled (void);
 
-struct mu_gsasl_module_data mu_gsasl_module_data;
-
-#ifdef USE_GSASL
+#ifdef WITH_GSASL
 #include <gsasl.h>
 
-int mu_gsasl_stream_create (mu_stream_t *stream, mu_stream_t transport,
-			    Gsasl_session_ctx *ctx,
-			    int flags);
+int gsasl_encoder_stream (mu_stream_t *pstr, mu_stream_t transport,
+			  Gsasl_session *ctx, int flags);
+int gsasl_decoder_stream (mu_stream_t *pstr, mu_stream_t transport,
+			  Gsasl_session *ctx, int flags);
 
 #endif
 
