@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999-2001, 2004-2007, 2009-2012, 2014-2016 Free
-   Software Foundation, Inc.
+   Copyright (C) 1999-2020 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -732,18 +731,6 @@ mu_mailer_set_stream (mu_mailer_t mailer, mu_stream_t stream)
   return 0;
 }
 
-int
-mu_mailer_get_stream (mu_mailer_t mailer, mu_stream_t * pstream)
-{
-  /* FIXME: Deprecation warning */
-  if (mailer == NULL)
-    return EINVAL;
-  if (pstream == NULL)
-    return MU_ERR_OUT_PTR_NULL;
-  *pstream = mailer->stream;
-  return 0;
-}
-
 int
 mu_mailer_get_streamref (mu_mailer_t mailer, mu_stream_t * pstream)
 {
