@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2005, 2007-2012, 2014-2016 Free Software Foundation,
-   Inc.
+   Copyright (C) 2005-2020 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -442,9 +441,9 @@ _construct_attr_array (size_t *pargc, char ***pargv)
   for (i = 0, mu_iterator_first (itr); !mu_iterator_is_done (itr);
        mu_iterator_next (itr), i++)
     {
-      char **str;
+      char *str;
       mu_iterator_current (itr, (void**) &str); 
-      if ((argv[i] = strdup (*str)) == NULL)
+      if ((argv[i] = strdup (str)) == NULL)
         {
           mu_argcv_free (i, argv);
           return ENOMEM;
@@ -538,8 +537,6 @@ _mu_entry_to_auth_data (LDAP *ld, LDAPMessage *msg,
   
   rc = ldap_get_dn_ber (ld, msg, &ber, &bv);
   ufn = ldap_dn2ufn (bv.bv_val);
-  /* FIXME: Use debug or diag functions */
-  mu_error ("INFO: %s", ufn);
   ldap_memfree (ufn);
   
   mu_assoc_get_iterator (ldap_param.field_map, &itr);
@@ -547,17 +544,15 @@ _mu_entry_to_auth_data (LDAP *ld, LDAPMessage *msg,
        mu_iterator_next (itr))
     {
       char *key;
-      char **pattr;
       char *attr;
       struct berval **values;
       
-      mu_iterator_current_kv (itr, (const void **)&key, (void**) &pattr);
-      attr = *pattr;
+      mu_iterator_current_kv (itr, (const void **)&key, (void**) &attr);
       values = ldap_get_values_len (ld, msg, attr);
       if (!values || !values[0])
 	{
 	  mu_error ("LDAP field `%s' (`%s') has NULL value",
-		    key, *pattr);
+		    key, attr);
 	  _free_partial_auth_data (&d);
 	  return MU_ERR_READ;
 	}
@@ -600,38 +595,34 @@ _mu_ldap_search (LDAP *ld, const char *filter_pat, const char *key,
   size_t nattrs;
   LDAPMessage *res, *msg;
   ber_int_t msgid;
-  const char *env[3];
-  struct mu_wordsplit ws;
-
+  char *filter_str;
+  
   rc = _construct_attr_array (&nattrs, &attrs);
   if (rc)
     return rc;
 
-  env[0] = "user";
-  env[1] = key;
-  env[2] = NULL;
-
-  ws.ws_env = env;
-  if (mu_wordsplit (filter_pat, &ws,
-		    MU_WRDSF_NOSPLIT | MU_WRDSF_NOCMD |
-		    MU_WRDSF_ENV | MU_WRDSF_ENV_KV))
+  rc = mu_str_vexpand (&filter_str, filter_pat, "user", key, NULL);
+  if (rc)
     {
-      mu_error (_("cannot expand line `%s': %s"), filter_pat,
-		mu_wordsplit_strerror (&ws));
-      return MU_ERR_FAILURE;
-    }
-  else if (ws.ws_wordc == 0)
-    {
-      mu_error (_("expanding %s yields empty string"), filter_pat);
-      mu_wordsplit_free (&ws);
       mu_argcv_free (nattrs, attrs);
-      return MU_ERR_FAILURE;
+      if (rc == MU_ERR_FAILURE)
+	{
+	  mu_error (_("cannot expand line `%s': %s"), filter_pat,
+		    filter_str);
+	  free (filter_str);
+	}
+      else
+	{
+	  mu_error (_("cannot expand line `%s': %s"), filter_pat,
+		    mu_strerror (rc));
+	}
+      return rc;
     }
   
   rc = ldap_search_ext (ld, ldap_param.base, LDAP_SCOPE_SUBTREE,
-			ws.ws_wordv[0], attrs, 0,
+			filter_str, attrs, 0,
 			NULL, NULL, NULL, -1, &msgid);
-  mu_wordsplit_free (&ws);
+  free (filter_str);
   mu_argcv_free (nattrs, attrs);
 
   if (rc != LDAP_SUCCESS)
@@ -643,7 +634,7 @@ _mu_ldap_search (LDAP *ld, const char *filter_pat, const char *key,
 	return MU_ERR_FAILURE;
     }
 
-  rc = ldap_result (ld, msgid, LDAP_MSG_ALL, NULL, &res );
+  rc = ldap_result (ld, msgid, LDAP_MSG_ALL, NULL, &res);
   if (rc < 0)
     {
       mu_error ("ldap_result failed");
