@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2004, 2005, 2006, 2007, 2008, 2009,
-   2010 Free Software Foundation, Inc.
+   Copyright (C) 1999-2020 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -13,9 +12,8 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General
-   Public License along with this library; if not, write to the
-   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-   Boston, MA 02110-1301 USA */
+   Public License along with this library.  If not, see
+   <http://www.gnu.org/licenses/>. */
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -40,9 +38,9 @@
 #include <mailutils/errno.h>
 #include <mailutils/progmailer.h>
 
-#include <url0.h>
-#include <mailer0.h>
-#include <registrar0.h>
+#include <mailutils/sys/url.h>
+#include <mailutils/sys/mailer.h>
+#include <mailutils/sys/registrar.h>
 
 static void sendmail_destroy (mu_mailer_t);
 static int sendmail_open (mu_mailer_t, int);
@@ -54,12 +52,7 @@ static int sendmail_send_message (mu_mailer_t, mu_message_t, mu_address_t,
 static int
 _url_sendmail_init (mu_url_t url)
 {
-  /* not valid in a sendmail url */
-  if (url->user || url->secret || url->auth || url->qargc
-      || url->host || url->port)
-    return EINVAL;
-
-  if (url->path == 0)
+  if (url->path == NULL)
     if ((url->path = strdup (PATH_SENDMAIL)) == 0)
       return ENOMEM;
 
@@ -115,9 +108,8 @@ sendmail_open (mu_mailer_t mailer, int flags)
 
   if (access (path, X_OK) == -1)
     return errno;
-  mu_progmailer_set_debug (pm, mailer->debug);
   status = mu_progmailer_set_command (pm, path);
-  MU_DEBUG1 (mailer->debug, MU_DEBUG_TRACE, "sendmail (%s)\n", path);
+  mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_TRACE, ("sendmail binary: %s", path));
   return status;
 }
 
@@ -161,19 +153,19 @@ sendmail_send_message (mu_mailer_t mailer, mu_message_t msg, mu_address_t from,
     {
       if ((status = mu_address_sget_email (from, 1, &emailfrom)) != 0)
 	{
-	  MU_DEBUG1 (mailer->debug, MU_DEBUG_ERROR,
-		     "cannot get recipient email: %s\n",
-		     mu_strerror (status));
+	  mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_ERROR,
+		    ("cannot get recipient email: %s",
+		     mu_strerror (status)));
 	  return status;
 	}
 
       if (!emailfrom)
 	{
 	  /* the address wasn't fully qualified, choke (for now) */
-	  MU_DEBUG1 (mailer->debug, MU_DEBUG_TRACE,
-		     "envelope from (%s) not fully qualifed\n",
-		     emailfrom);
-	  return MU_ERR_BAD_822_FORMAT;
+	  mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_TRACE,
+		    ("envelope from (%s) not fully qualifed\n",
+		     emailfrom));
+	  return MU_ERR_INVALID_EMAIL;
 	}
 
       argc += 2;		/* -f from */
@@ -187,8 +179,7 @@ sendmail_send_message (mu_mailer_t mailer, mu_message_t msg, mu_address_t from,
 
       if (tocount == 0)
 	{
-	  MU_DEBUG (mailer->debug, MU_DEBUG_TRACE,
-		    "missing recipients\n");
+	  mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_TRACE, ("missing recipients"));
 	  return MU_ERR_NOENT;
 	}
       
@@ -235,26 +226,25 @@ sendmail_send_message (mu_mailer_t mailer, mu_message_t msg, mu_address_t from,
 	  if ((status = mu_address_sget_email (to, i, &email)) != 0)
 	    {
 	      free (argvec);
-	      MU_DEBUG2 (mailer->debug, MU_DEBUG_ERROR,
-			 "cannot get email of recipient #%lu: %s\n",
-			 (unsigned long) i, mu_strerror (status));
+	      mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_ERROR,
+			("cannot get email of recipient #%lu: %s",
+			 (unsigned long) i, mu_strerror (status)));
 	      return status;
 	    }
 	  
 	  if (!email)
 	    {
-	      MU_DEBUG1 (mailer->debug, MU_DEBUG_TRACE,
-			 "envelope to (%s) not fully qualifed\n",
-			 email);
+	      mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_TRACE,
+			("envelope to (%s) not fully qualifed",
+			 email));
 	      free (argvec);
-	      return MU_ERR_BAD_822_FORMAT;
+	      return MU_ERR_INVALID_EMAIL;
 	    }
 	  argvec[argc++] = email;
 	}
     }
   argvec[argc] = NULL;
   
-  mu_progmailer_set_debug (pm, mailer->debug);
   status = mu_progmailer_open (pm, (char**) argvec);
   if (status == 0)
     {
@@ -263,9 +253,8 @@ sendmail_send_message (mu_mailer_t mailer, mu_message_t msg, mu_address_t from,
 	  mu_observable_notify (mailer->observable, MU_EVT_MAILER_MESSAGE_SENT,
 				msg);
       else
-	MU_DEBUG1 (mailer->debug, MU_DEBUG_ERROR,
-		   "progmailer error: %s\n",
-		   mu_strerror (status));
+	mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_ERROR,
+		  ("progmailer error: %s", mu_strerror (status)));
     }
   
   free (argvec);
@@ -276,6 +265,10 @@ static struct _mu_record _sendmail_record =
 {
   MU_SENDMAIL_PRIO,
   MU_SENDMAIL_SCHEME,
+  MU_RECORD_DEFAULT,
+  MU_URL_SCHEME | MU_URL_PATH,
+  MU_URL_SCHEME, /* Nothing is required in the URL, except scheme. 
+                    Missing path means using PATH_SENDMAIL. */
   _url_sendmail_init,    /* url init.  */
   _mu_mailer_mailbox_init,     /* Mailbox entry.  */
   _mu_mailer_sendmail_init, /* Mailer entry.  */
@@ -293,6 +286,6 @@ mu_record_t mu_sendmail_record = &_sendmail_record;
 
 #else
 #include <stdio.h>
-#include <registrar0.h>
+#include <mailutils/sys/registrar.h>
 mu_record_t mu_sendmail_record = NULL;
 #endif
