@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2004-2019 Free Software Foundation, Inc.
+   Copyright (C) 2004-2020 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -189,7 +189,7 @@ nntp_folder_destroy (mu_folder_t folder)
 
 static int
 nntp_folder_list (mu_folder_t folder, const char *ref, void *pat, int flags,
-		  size_t max_level, mu_list_t flist,
+		  size_t max_depth, mu_list_t flist,
 		  mu_folder_enumerate_fp efp, void *edp)
 {
   return ENOTSUP;
