@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2005-2012, 2014-2017 Free Software Foundation, Inc.
+   Copyright (C) 2005-2020 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -75,8 +75,14 @@ int
 burst_mime (mu_message_t msg)
 {
   size_t i, nparts;
+  int rc;
   
-  mu_message_get_num_parts (msg, &nparts);
+  rc = mu_message_get_num_parts (msg, &nparts);
+  if (rc)
+    {
+      mu_diag_funcall (MU_DIAG_ERR, "mu_message_get_num_parts", NULL, rc);
+      return rc;
+    }
 
   for (i = 1; i <= nparts; i++)
     {
@@ -650,10 +656,12 @@ main (int argc, char **argv)
   int rc;
   mu_mailbox_t mbox;
   mu_msgset_t msgset;
-  const char *tempfolder = mh_global_profile_get ("Temp-Folder", ".temp");
-  
+  const char *tempfolder = NULL;
+
   mh_getopt (&argc, &argv, options, MH_GETOPT_DEFAULT_FOLDER,
 	     args_doc, prog_doc, NULL);
+  if (!tempfolder)
+    tempfolder = mh_global_profile_get ("Temp-Folder", ".temp");
   if (eb_min_length == 0)
     eb_min_length = 1;
   
